# Copyright 2019 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0
import traceback
from contextlib import suppress
from functools import partial
from types import SimpleNamespace

from hissp.reader import Parser
from hissp.reader import transpile


def repl(macros=None):
    parser = Parser()
    if not macros:
        with suppress(FileNotFoundError):
            transpile("hissp", "basic")
        from hissp import basic

        macros = basic
    parser.compiler.ns["_macro_"] = SimpleNamespace(**vars(macros._macro_))
    while True:
        try:
            try:
                line = input("\n#> ")
            except EOFError:
                print("Ssss.")
                break
            buffer = _get_more(line)
            forms = parser.reads("\n".join(buffer))
            evaluate(forms, parser)
        except SystemExit:
            print("Lissp REPL Exit.")
            raise
        except BaseException as be:
            traceback.print_exception(type(be), be, be.__traceback__.tb_next)


def evaluate(forms, parser):
    code = parser.compiler.compile(forms)
    print(">>>", code.replace("\n", "\n... "))
    bytecode = compile(code, "<repl>", "single")
    exec(bytecode, parser.compiler.ns)
    return code


def _get_more(line):
    buffer = [line]
    if "(" in line or '"' in line or ";" in line:
        buffer.extend(iter(partial(input, "#.."), ""))
    return buffer



# Copyright 2019 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0

import ast
import builtins
import os
import re
from contextlib import contextmanager, nullcontext
from functools import reduce
from importlib import import_module, resources
from itertools import chain
from pathlib import Path, PurePath
from pprint import pprint
from types import ModuleType
from typing import Any, Iterable, Iterator, NewType, Tuple, Union
from unittest.mock import ANY

from hissp.compiler import Compiler, readerless
from hissp.munger import munge

TOKENS = re.compile(
    r"""(?x)
 (?P<open>\()
|(?P<close>\))
|(?P<string>
  " # Open quote.
    (?:|[^"\\]  # Any non-magic character.
       |\\(?:.|\n)  # Backslash only if paired, including with newline.
    )*  # Zero or more times.
  " # Close quote.
 )
|(?P<comment>;.*)
|(?P<whitespace>[\n ]+)  # Tabs are not allowed outside of strings.
|(?P<macro>
   ,@
  |['`,]
   # Ends in ``#``, but not bytes, dict, set, list, str.
  |(?:[Bb](?!')
     |[^ \n"(){}[\]#Bb]
     )[^ \n"(){}[\]#]*[#])
|(?P<symbol>[^ \n"()]+)
"""
)

Token = NewType("Token", Tuple[str, str])

DROP = object()


def lex(code: str, file: str = "<?>") -> Iterator[Token]:
    pos = 0
    while pos < len(code):
        match = TOKENS.match(code, pos)
        if match is None:
            good = code[0:pos].split("\n")
            line = len(good)
            column = len(good[-1])
            raise SyntaxError("Unexpected token", (file, line, column, code))
        assert match.end() > pos, match.groups()
        pos = match.end()
        yield Token((match.lastgroup, match.group()))


class _Unquote(tuple):
    def __repr__(self):
        return f"_Unquote{super().__repr__()}"


def gensym_counter(count=[0]):
    count[0] += 1
    return count[0]


class Parser:
    def __init__(
        self, qualname="_repl", ns=None, verbose=False, evaluate=False, filename="<?>"
    ):
        self.qualname = qualname
        self.ns = ns or {"__name__": "<compiler>"}
        self.compiler = Compiler(self.qualname, self.ns, evaluate)
        self.verbose = verbose
        self.filename = filename
        self.reinit()

    def reinit(self):
        self.gensym_stack = []
        self.depth = 0

    def parse(self, tokens: Iterator[Token]) -> Iterator:
        return (form for form in self._parse(tokens) if form is not DROP)

    def _parse(self, tokens: Iterator[Token]) -> Iterator:
        for k, v in tokens:
            if k == "open":
                depth = self.depth
                self.depth += 1
                yield (*self.parse(tokens),)
                if self.depth != depth:
                    raise SyntaxError("Unclosed '('.")
            elif k == "close":
                self.depth -= 1
                if self.depth < 0:
                    raise SyntaxError("Unopened ')'.")
                return
            elif k == "string":
                yield "quote", ast.literal_eval(
                    v.replace("\\\n", "").replace("\n", r"\n")
                ), {":str": True}
            elif k in {"comment", "whitespace"}:
                continue
            elif k == "macro":
                with {
                    "`": self.gensym_context,
                    ",": self.unquote_context,
                    ",@": self.unquote_context,
                }.get(v, nullcontext)():
                    form = next(self.parse(tokens))
                    yield self.parse_macro(v, form)
            elif k == "symbol":
                try:
                    yield ast.literal_eval(v)
                except (ValueError, SyntaxError):
                    yield munge(v)
            else:
                assert False, "unknown token: " + repr(k)
        if self.depth:
            SyntaxError("Ran out of tokens before completing form.")

    def parse_macro(self, tag: str, form):
        if tag == "'":
            return "quote", form
        if tag == "`":
            return self.template(form)
        if tag == ",":
            return _Unquote([":?", form])
        if tag == ",@":
            return _Unquote([":*", form])

        assert tag.endswith("#")
        tag = tag[:-1]
        if tag == "_":
            return DROP
        if tag == "$":
            return self.gensym(form)
        if tag == ".":
            return eval(readerless(form), {})
        if ".." in tag and not tag.startswith(".."):
            module, function = tag.split("..", 1)
            function = munge(function)
            if is_string(form):
                form = form[1]
            return reduce(getattr, function.split("."), import_module(module))(form)
        raise ValueError(f"Unknown reader macro {tag}")

    def template(self, form):
        case = type(form)
        if case is tuple and form:
            if is_string(form):
                return "quote", form
            return (
                ("lambda", (":", ":*", "xAUTO0_"), "xAUTO0_"),
                ":",
                *chain(*self._template(form)),
            )
        if case is str and not form.startswith(":"):
            return "quote", self.qualify(form)
        if case is _Unquote and form[0] == ":?":
            return form[1]
        return form

    def _template(self, forms: Iterable) -> Iterable[Tuple[str, Any]]:
        for form in forms:
            case = type(form)
            if case is str and not form.startswith(":"):
                yield ":?", ("quote", self.qualify(form))
            elif case is _Unquote:
                yield form
            elif case is tuple:
                yield ":?", self.template(form)
            else:
                yield ":?", form

    def qualify(self, symbol: str) -> str:
        if symbol in {e for e in dir(builtins) if not e.startswith("_")}:
            return f"builtins..{symbol}"
        if re.search(r"\.\.|^\.|^quote$|^lambda$|xAUTO\d+_$", symbol):
            return symbol
        if symbol in vars(self.ns.get("_macro_", lambda: ())):
            return f"{self.qualname}.._macro_.{symbol}"
        return f"{self.qualname}..{symbol}"

    def reads(self, code: str) -> Iterable:
        res = self.parse(lex(code, self.filename))
        self.reinit()
        if self.verbose:
            res = list(res)
            pprint(res)
        return res

    def compile(self, code: str) -> str:
        hissp = self.reads(code)
        return self.compiler.compile(hissp)

    def gensym(self, form: str):
        return f"_{munge(form)}xAUTO{self.gensym_stack[-1]}_"

    @contextmanager
    def gensym_context(self):
        self.gensym_stack.append(gensym_counter())
        try:
            yield
        finally:
            self.gensym_stack.pop()

    @contextmanager
    def unquote_context(self):
        gensym_number = self.gensym_stack.pop()
        try:
            yield
        finally:
            self.gensym_stack.append(gensym_number)


def is_string(form):
    return form == ("quote", ANY, ANY) and form[2].get(":str")


def transpile(package: resources.Package, *modules: Union[str, PurePath]):
    for module in modules:
        transpile_module(package, module + ".lissp")


def transpile_module(
    package: resources.Package,
    resource: Union[str, PurePath],
    out: Union[None, str, bytes, Path] = None,
):
    code = resources.read_text(package, resource)
    path: Path
    with resources.path(package, resource) as path:
        out = out or path.with_suffix(".py")
        if isinstance(package, ModuleType):
            package = package.__package__
        if isinstance(package, os.PathLike):
            resource = resource.stem
        qualname = f"{package}.{resource.split('.')[0]}"
        with open(out, "w") as f:
            print("writing to", out)
            f.write(Parser(qualname, evaluate=True, filename=str(out)).compile(code))



# Copyright 2019 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0

import re
from contextlib import suppress
from typing import Dict, Hashable, Mapping, Match, TypeVar

import unicodedata


def munge(s: str) -> str:
    if s.startswith(":") or s.isidentifier():
        return s
    return "".join(map(x_quote, s))


TO_NAME = {
    "~": "xTILDE_",
    "`": "xGRAVE_",
    "!": "xBANG_",
    "@": "xAT_",
    "#": "xHASH_",
    "$": "xDOLLAR_",
    "%": "xPERCENT_",
    "^": "xCARET_",
    "&": "xET_",
    "*": "xSTAR_",
    "-": "xH_",  # Hyphen
    "+": "xPLUS_",
    "=": "xEQ_",
    "|": "xBAR_",
    "\\": "xBSLASH_",
    ":": "xCOLON_",
    "'": "xQUOTE_",
    "<": "xLT_",  # Less Than or LefT.
    ",": "xCOMMA_",
    ">": "xGT_",  # Greater Than or riGhT.
    "?": "xQUERY_",
    "/": "xSLASH_",
    " ": "xSPACE_",
}
X_NAME = {ord(k): ord(v) for k, v in {" ": "x", "-": "h"}.items()}


def x_quote(c: str) -> str:
    return (
        TO_NAME.get(c)
        or unicodedata.category(c) == "Sm"
        and f"x{unicodedata.name(c).translate(X_NAME)}_"
        or c
    )


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


def reversed_1to1(mapping: Mapping[K, V]) -> Dict[V, K]:
    result = {v: k for k, v in mapping.items()}
    assert len(mapping) == len(result)
    return result


LOOKUP_NAME = reversed_1to1(TO_NAME)
UN_X_NAME = reversed_1to1(X_NAME)


def un_x_quote(match: Match[str]) -> str:
    with suppress(KeyError):
        return LOOKUP_NAME.get(match.group()) or unicodedata.lookup(
            match.group(1).translate(UN_X_NAME)
        )
    return match.group()


def demunge(s: str) -> str:
    return re.sub("x([0-9A-Zhx]+?)_", un_x_quote, s)



#!/usr/bin/env python3
import setuptools

with open("README.md") as f:
    long_description = f.read()

setuptools.setup(
    name="hissp",
    version="0.1.0",
    description="It's Python with a Lissp.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Matthew Egan Odendahl",
    author_email="hissp.gilch@xoxy.net",
    license="Apache-2.0",
    url="https://github.com/gilch/hissp",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Compilers",
        "Topic :: Software Development :: Libraries",
    ],
    keywords="lisp macro metaprogramming compiler DSL AST transpiler emacs clojure scheme",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.7",
    entry_points={"console_scripts": ["hissp=hissp.__main__:repl"]},
)



__import__('operator').setitem(
  globals(),
  '_macro_',
  __import__('types').ModuleType(
    '_macro_'))

setattr(
  _macro_,
  'requirexH_as',
  (lambda macro,name:
    (lambda *xAUTO0_:xAUTO0_)(
      'builtins..setattr',
      (lambda *xAUTO0_:xAUTO0_)(
        '.setdefault',
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..globals'),
        (lambda *xAUTO0_:xAUTO0_)(
          'quote',
          '_macro_'),
        (lambda *xAUTO0_:xAUTO0_)(
          'types..ModuleType',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            'hissp.basic.._macro_'))),
      (lambda *xAUTO0_:xAUTO0_)(
        'quote',
        name),
      macro)))

setattr(
  _macro_.requirexH_as,
  '__doc__',
  "\nRequire a macro callable as an unqualified name.\n\nAdds the macro to the current module's _macro_ space under the specified\nname, which makes it available unqualified in this module. (And\navailable qualified with this module's name.)\n")

setattr(
  _macro_.requirexH_as,
  '__qualname__',
  '_macro_.requirexH_as')

# requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'defmacro',
  (lambda name,parameters,*body:
    (lambda *xAUTO0_:xAUTO0_)(
      'hissp.basic.._macro_.requirexH_as',
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        parameters,
        *body),
      name)))

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'let',
  (lambda pairs,*body:
    (lambda *xAUTO0_:xAUTO0_)(
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (lambda *xAUTO0_:xAUTO0_)(
          ':',
          *pairs),
        *body))))

setattr(
  _macro_.let,
  '__doc__',
  '\nCreates locals. Pairs are implied. Locals are not in scope until the body.\n')

setattr(
  _macro_.let,
  '__qualname__',
  '_macro_.let')

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'defmacro',
  (lambda name,parameters,docstring,*body:
    (lambda *xAUTO0_:xAUTO0_)(
      'hissp.basic.._macro_.requirexH_as',
      (lambda *xAUTO0_:xAUTO0_)(
        'hissp.basic.._macro_.let',
        (lambda *xAUTO0_:xAUTO0_)(
          '_macroxAUTO4_',
          (lambda *xAUTO0_:xAUTO0_)(
            'lambda',
            parameters,
            *body)),
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..setattr',
          '_macroxAUTO4_',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            '__doc__'),
          docstring),
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..setattr',
          '_macroxAUTO4_',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            '__qualname__'),
          (lambda *xAUTO0_:xAUTO0_)(
            '.join',
            ('quote', '.', {':str': True}),
            (lambda *xAUTO0_:xAUTO0_)(
              'quote',
              (lambda *xAUTO0_:xAUTO0_)(
                '_macro_',
                name)))),
        '_macroxAUTO4_'),
      name)))

setattr(
  _macro_.defmacro,
  '__doc__',
  '\nCreates a new macro for the current module.\n\nThe docstring argument is required, but can be None.\n')

setattr(
  _macro_.defmacro,
  '__qualname__',
  '_macro_.defmacro')

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'cascade',
  # hissp.basic.._macro_.let
  (lambda _macroxAUTO4_=(lambda thing,*calls:
    # let
    (lambda thingxH_sym='_thingxAUTO5_':
      (lambda *xAUTO0_:xAUTO0_)(
        (lambda *xAUTO0_:xAUTO0_)(
          'lambda',
          (lambda *xAUTO0_:xAUTO0_)(
            ':',
            thingxH_sym,
            thing),
          *map(
            (lambda call:
              (lambda *xAUTO0_:xAUTO0_)(
                __import__('operator').getitem(
                  call,
                  (0)),
                thingxH_sym,
                *__import__('operator').getitem(
                  call,
                  slice(
                    (1),
                    None)))),
            calls),
          thingxH_sym)))()):(
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__doc__',
      'Call multiple methods on one object.\n\n  Evaluates the given thing then uses it as the first argument to a\n  sequence of calls. Used for initialization. Evaluates to the thing.\n  '),
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__qualname__',
      '.'.join(
        ('_macro_', 'cascade'))),
    _macroxAUTO4_)[-1])())

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'define',
  # hissp.basic.._macro_.let
  (lambda _macroxAUTO4_=(lambda name,value:
    (lambda *xAUTO0_:xAUTO0_)(
      'operator..setitem',
      (lambda *xAUTO0_:xAUTO0_)(
        'builtins..globals'),
      (lambda *xAUTO0_:xAUTO0_)(
        'quote',
        name),
      value)):(
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__doc__',
      'Assigns a global in the current module.'),
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__qualname__',
      '.'.join(
        ('_macro_', 'define'))),
    _macroxAUTO4_)[-1])())

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'ifxH_else',
  # hissp.basic.._macro_.let
  (lambda _macroxAUTO4_=(lambda test,then,otherwise:
    (lambda *xAUTO0_:xAUTO0_)(
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (lambda *xAUTO0_:xAUTO0_)(
          'test',
          ':',
          ':*',
          'thenxH_else'),
        (lambda *xAUTO0_:xAUTO0_)(
          (lambda *xAUTO0_:xAUTO0_)(
            'operator..getitem',
            'thenxH_else',
            (lambda *xAUTO0_:xAUTO0_)(
              'operator..not_',
              'test')))),
      test,
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (),
        then),
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (),
        otherwise))):(
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__doc__',
      "Basic ternary branching construct.\n\n  Like Python's conditional expressions, the else-clause is required.\n  "),
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__qualname__',
      '.'.join(
        ('_macro_', 'ifxH_else'))),
    _macroxAUTO4_)[-1])())

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'progn',
  # hissp.basic.._macro_.let
  (lambda _macroxAUTO4_=(lambda *body:
    (lambda *xAUTO0_:xAUTO0_)(
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (),
        *body))):(
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__doc__',
      'Evaluates each form in sequence for side effects.\n\n  Evaluates to the same value as its last form (or ``()`` if empty).\n  '),
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__qualname__',
      '.'.join(
        ('_macro_', 'progn'))),
    _macroxAUTO4_)[-1])())

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'fromxH_require',
  # hissp.basic.._macro_.let
  (lambda _macroxAUTO4_=(lambda *packagexPLUS_macros:
    (lambda *xAUTO0_:xAUTO0_)(
      'hissp.basic.._macro_.progn',
      (lambda *xAUTO0_:xAUTO0_)(
        '.setdefault',
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..globals'),
        (lambda *xAUTO0_:xAUTO0_)(
          'quote',
          '_macro_'),
        (lambda *xAUTO0_:xAUTO0_)(
          'types..ModuleType',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            'hissp.basic.._macro_'))),
      *__import__('itertools').starmap(
        (lambda package,*macros:
          (lambda *xAUTO0_:xAUTO0_)(
            'hissp.basic.._macro_.progn',
            *map(
              (lambda macro:
                (lambda *xAUTO0_:xAUTO0_)(
                  'builtins..setattr',
                  '_macro_',
                  (lambda *xAUTO0_:xAUTO0_)(
                    'quote',
                    macro),
                  '{}.._macro_.{}'.format(
                    package,
                    macro))),
              macros))),
        packagexPLUS_macros))):(
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__doc__',
      'Adds macros for the current module from ``package``\n\n   For example::\n\n    (from-require (foo.package spammacro eggsmacro)\n                  (bar.package baconmacro bannanamacro))\n\n  '),
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__qualname__',
      '.'.join(
        ('_macro_', 'fromxH_require'))),
    _macroxAUTO4_)[-1])())

# defmacro
# hissp.basic.._macro_.requirexH_as
__import__('builtins').setattr(
  __import__('builtins').globals().setdefault(
    '_macro_',
    __import__('types').ModuleType(
      'hissp.basic.._macro_')),
  'deftype',
  # hissp.basic.._macro_.let
  (lambda _macroxAUTO4_=(lambda name,bases,*body:
    (lambda *xAUTO0_:xAUTO0_)(
      'hissp.basic.._macro_.define',
      name,
      (lambda *xAUTO0_:xAUTO0_)(
        'builtins..type',
        (lambda *xAUTO0_:xAUTO0_)(
          'quote',
          name),
        (lambda *xAUTO0_:xAUTO0_)(
          (lambda *xAUTO0_:xAUTO0_)(
            'lambda',
            (lambda *xAUTO0_:xAUTO0_)(
              ':',
              ':*',
              'xAUTO0_'),
            'xAUTO0_'),
          *bases),
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..dict',
          ':',
          *body)))):(
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__doc__',
      'Defines a type (class) in the current module.\n\n  Key-value pairs are implied in the body.\n  '),
    __import__('builtins').setattr(
      _macroxAUTO4_,
      '__qualname__',
      '.'.join(
        ('_macro_', 'deftype'))),
    _macroxAUTO4_)[-1])())

# define
__import__('operator').setitem(
  __import__('builtins').globals(),
  '__doc__',
  "\nHissp's basic macros.\n\nThese are automatically made available as unqualified macros in the\nbasic REPL. To use them in a Hissp module, either use the\nfully-qualified names, or add them to the module's _macro_'s.\n\nFor example::\n\n  (hissp.basic.._macro_.from-require\n   (hissp.basic cascade if-else let progn))\n\n")


# Copyright 2019 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0

import ast
import pickle
import pickletools
import re
from contextlib import contextmanager, suppress
from contextvars import ContextVar
from functools import wraps
from itertools import chain, takewhile
from pprint import pformat
from typing import Iterable, Tuple, TypeVar

PAIR_WORDS = {":*": "*", ":**": "**", ":?": ""}
# Module Macro container
MACROS = "_macro_"
# Macro from foreign module foo.bar.._macro_.baz
MACRO = f"..{MACROS}."

NUMBER = frozenset({int, float, complex})

# Sometimes macros need the current ns when expanding,
# instead of its defining ns.
# Rather than pass in an implicit argument, it's available here.
# readerless() uses this automatically.
NS = ContextVar("NS", default=None)


class CompileError(SyntaxError):
    pass


def trace(method):
    @wraps(method)
    def tracer(self, expr):
        try:
            return method(self, expr)
        except CompileError as e:
            e.msg = f"\nIn compile {method.__name__}:\n{pformat(expr)}\n" + e.msg
            raise e
        except Exception as e:
            raise CompileError(
                f"\nIn compile {method.__name__}:\n{pformat(expr)}\n\n{type(e).__name__}: {e}"
            ) from e

    return tracer


class Compiler:
    """
    The Hissp compiler.
    """

    def __init__(self, qualname="_repl", ns=None, evaluate=True):
        self.qualname = qualname
        self.ns = ns or {"__name__": "<compiler>"}
        self.evaluate = evaluate

    def compile(self, forms: Iterable) -> str:
        result = []
        for form in forms:
            form = self.form(form)
            self.eval(form)
            result.append(form)
        return "\n\n".join(result)

    def eval(self, form):
        if not self.evaluate:
            return
        try:
            eval(compile(form, "<Hissp>", "eval"), self.ns)
        except Exception as e:
            print("Exception when evaluating form:")
            print(form)
            raise CompileError(f"\n{form!r}") from e

    def form(self, form) -> str:
        """
        Translate Hissp form to the equivalent Python code as a string.
        """
        if type(form) is tuple and form:
            return self.tuple(form)
        if type(form) is str and not form.startswith(":"):
            return self.symbol(form)
        return self.quoted(form)

    def tuple(self, form: tuple) -> str:
        """Calls, macros, special forms."""
        head, *tail = form
        if type(head) is str:
            return self.special(form)
        return self.call(form)

    def special(self, form: tuple) -> str:
        """Try to compile as special form, else self.macro()."""
        if form[0] == "quote":
            return self.quoted(form[1])
        if form[0] == "lambda":
            return self.function(form)
        return self.invocation(form)

    @trace
    def invocation(self, form: tuple) -> str:
        """Try to compile as macro, else normal call."""
        head, *tail = form
        parts = head.split(MACRO, 1)
        with self.macro_context():
            if parts[0] == self.qualname:
                # Local qualified macro. Recursive macros might need it.
                return f"# {head}\n" + self.form(vars(self.ns[MACROS])[parts[1]](*tail))
            try:  # Is it a local unqualified macro?
                macro = vars(self.ns[MACROS])[head]
            except LookupError:  # Nope.
                pass
            else:  # Yes.
                return f"# {head}\n" + self.form(macro(*tail))
            if MACRO in head:  # Qualified macro, not local.
                return f"# {head}\n" + self.form(eval(self.symbol(head))(*tail))
        return self.call(form)

    def quoted(self, form) -> str:
        r"""
        Compile forms that evaluate to themselves.

        Emits a literal if possible, otherwise falls back to pickle:

        >>> readerless(-4.2j)
        '((-0-4.2j))'
        >>> print(readerless(float('nan')))
        __import__('pickle').loads(  # nan
            b'\x80\x03G\x7f\xf8\x00\x00\x00\x00\x00\x00.'
        )
        >>> readerless([{'foo':2},(),1j,2.0,{3}])
        "[{'foo': 2}, (), 1j, 2.0, {3}]"
        >>> spam = []
        >>> spam.append(spam)
        >>> print(readerless(spam))
        __import__('pickle').loads(  # [[...]]
            b'\x80\x03]q\x00h\x00a.'
        )

        """
        # Number literals may need (). E.g. (1).real
        literal = f"({form!r})" if type(form) in NUMBER else repr(form)
        with suppress(ValueError):
            if ast.literal_eval(literal) == form:
                return literal
        # Wasn't made with ast.literal_eval(). Fall back to pickle.
        return self.pickle(form)

    @trace
    def pickle(self, form) -> str:
        """The final fallback for self.quoted()."""
        dumps = pickletools.optimize(pickle.dumps(form))
        return f"__import__('pickle').loads(  # {form!r}\n    {dumps}\n)"

    def function(self, form: tuple) -> str:
        r"""
        Anonymous function special form.

        (lambda (<parameters>)
          <body>)

        The parameters tuple is divided into (<single> : <paired>)

        Parameter types are the same as Python's.
        For example,

        >>> readerless(
        ... ('lambda', ('a','b',
        ...         ':', 'e',1, 'f',2,
        ...         ':*','args', 'h',4, 'i',':?', 'j',1,
        ...         ':**','kwargs',),
        ...   42,),
        ... )
        '(lambda a,b,e=(1),f=(2),*args,h=(4),i,j=(1),**kwargs:(42))'

        The special keywords :* and :** designate the remainder of the
        positional and keyword parameters, respectively.
        Note this body has an implicit PROGN:

        >>> print(readerless(
        ... ('lambda', (':',':*','args',':**','kwargs',),
        ...   ('print','args',),
        ...   ('print','kwargs',),),
        ... ))
        (lambda *args,**kwargs:(
          print(
            args),
          print(
            kwargs))[-1])

        You can omit the right of a pair with ``:?``
        (except the final ``**kwargs``).
        Also note that the body can be empty.

        >>> readerless(
        ... ('lambda', (':','a',1, ':*',':?', 'b',':?', 'c',2,),),
        ... )
        '(lambda a=(1),*,b,c=(2):())'

        The ':' may be omitted if there are no paired parameters.

        >>> readerless(('lambda', ('a','b','c',':',),),)
        '(lambda a,b,c:())'
        >>> readerless(('lambda', ('a','b','c',),),)
        '(lambda a,b,c:())'
        >>> readerless(('lambda', (':',),),)
        '(lambda :())'
        >>> readerless(('lambda', (),),)
        '(lambda :())'

        ``:`` is required if there are any paired parameters, even if
        there are no single parameters:

        >>> readerless(('lambda', (':',':**','kwargs',),),)
        '(lambda **kwargs:())'

        """
        fn, parameters, *body = form
        assert fn == "lambda"
        return f"(lambda {','.join(self.parameters(parameters))}:{self.body(body)})"

    @trace
    def parameters(self, parameters: tuple) -> Iterable[str]:
        parameters = iter(parameters)
        yield from takewhile(lambda a: a != ":", parameters)
        for k, v in pairs(parameters):
            if k == ":*":
                yield "*" if v == ":?" else f"*{v}"
            elif k == ":**":
                yield f"**{v}"
            elif v == ":?":
                yield k
            else:
                yield f"{k}={self.form(v)}"

    @trace
    def body(self, body: list) -> str:
        if len(body) > 1:
            return f"({_join_args(*map(self.form, body))})[-1]"
        if not body:
            return "()"
        result = self.form(body[0])
        return ("\n" * ("\n" in result) + result).replace("\n", "\n  ")

    @trace
    def call(self, form: tuple) -> str:
        r"""
        Call form.

        Any tuple that is not quoted, empty, or a special form or macro is
        a runtime call.

        Like Python, it has three parts.
        (<callable> <args> : <kwargs>)
        For example:

        >>> print(readerless(
        ... ('print',1,2,3,
        ...          ':','sep',('quote',":",), 'end',('quote',"\n\n",),)
        ... ))
        print(
          (1),
          (2),
          (3),
          sep=':',
          end='\n\n')

        Either <args> or <kwargs> may be empty:

        >>> readerless(('foo',':',),)
        'foo()'
        >>> print(readerless(('foo','bar',':',),))
        foo(
          bar)
        >>> print(readerless(('foo',':','bar','baz',),))
        foo(
          bar=baz)

        The : is optional if the <kwargs> part is empty:

        >>> readerless(('foo',),)
        'foo()'
        >>> print(readerless(('foo','bar',),),)
        foo(
          bar)

        The <kwargs> part has implicit pairs; there must be an even number.

        Use the special keywords ``:*`` and ``:**`` for iterable and
        mapping unpacking:

        >>> print(readerless(
        ... ('print',':',':*',[1,2], 'a',3, ':*',[4], ':**',{'sep':':','end':'\n\n'},),
        ... ))
        print(
          *[1, 2],
          a=(3),
          *[4],
          **{'sep': ':', 'end': '\n\n'})

        Unlike other keywords, these can be repeated,
        but (as in Python) a '*' is not allowed to follow '**'.

        Method calls are similar to function calls.
        (.<method name> <object> <args> : <kwargs>)
        Like Clojure, a method on the first object is assumed if the
        function name starts with a dot:

        >>> readerless(('.conjugate', 1j,),)
        '(1j).conjugate()'
        >>> eval(_)
        -1j
        >>> readerless(('.decode', b'\xfffoo', ':', 'errors',('quote','ignore',),),)
        "b'\\xfffoo'.decode(\n  errors='ignore')"
        >>> eval(_)
        'foo'

        """
        form = iter(form)
        head = next(form)
        args = chain(
            map(self.form, takewhile(lambda a: a != ":", form)),
            (f"{(PAIR_WORDS.get(k, k+'='))}{self.form(v)}" for k, v in pairs(form)),
        )
        if type(head) is str and head.startswith("."):
            return "{}.{}({})".format(next(args), head[1:], _join_args(*args))
        return "{}({})".format(self.form(head), _join_args(*args))

    @trace
    def symbol(self, symbol: str) -> str:
        if re.search(r"^\.\.|[ ()]", symbol):  # Python injection?
            return symbol
        if ".." in symbol:
            parts = symbol.split("..", 1)
            if parts[0] == self.qualname:  # This module. No import required.
                chain = parts[1].split('.', 1)
                chain[0] = f"globals()[{self.quoted(chain[0])}]"  # Avoid local shadowing.
                return '.'.join(chain)
            return "__import__({0!r}{fromlist}).{1}".format(
                parts[0], parts[1], fromlist=",fromlist='?'" if "." in parts[0] else ""
            )
        return symbol

    @contextmanager
    def macro_context(self):
        token = NS.set(self.ns)
        try:
            yield
        finally:
            NS.reset(token)


def _join_args(*args):
    return (("\n" if args else "") + ",\n".join(args)).replace("\n", "\n  ")


T = TypeVar("T")


def pairs(it: Iterable[T]) -> Iterable[Tuple[T, T]]:
    it = iter(it)
    for k in it:
        yield k, next(it)


def readerless(form, ns=None):
    ns = ns or NS.get() or {"__name__": "<compiler>"}
    return Compiler(evaluate=False, ns=ns).compile([form])
